@@ -1,9 +1,10 @@
 /*
- * Copyright  2001-2004 The Apache Software Foundation
- *
- *  Licensed under the Apache License, Version 2.0 (the "License");
- *  you may not use this file except in compliance with the License.
- *  You may obtain a copy of the License at
+ *  Licensed to the Apache Software Foundation (ASF) under one or more
+ *  contributor license agreements.  See the NOTICE file distributed with
+ *  this work for additional information regarding copyright ownership.
+ *  The ASF licenses this file to You under the Apache License, Version 2.0
+ *  (the "License"); you may not use this file except in compliance with
+ *  the License.  You may obtain a copy of the License at
  *
  *      http://www.apache.org/licenses/LICENSE-2.0
  *
@@ -16,17 +17,16 @@
  */
 package org.apache.tools.ant.types;
 
-// java io classes
 import java.io.File;
 import java.io.FileInputStream;
-import java.io.IOException;
 import java.util.Enumeration;
 import java.util.Hashtable;
 import java.util.Properties;
 import java.util.Vector;
-import org.apache.tools.ant.BuildException;
-import org.apache.tools.ant.Project;
 
+import org.apache.tools.ant.Project;
+import org.apache.tools.ant.BuildException;
+import org.apache.tools.ant.util.FileUtils;
 
 /**
  * A set of filters to be applied to something.
@@ -37,64 +37,66 @@ import org.apache.tools.ant.Project;
 public class FilterSet extends DataType implements Cloneable {
 
     /**
-     * Individual filter component of filterset
+     * Individual filter component of filterset.
      *
      */
     public static class Filter {
-        /** Token which will be replaced in the filter operation */
+        // CheckStyle:VisibilityModifier OFF - bc
+        /** Token which will be replaced in the filter operation. */
         String token;
 
-        /** The value which will replace the token in the filtering operation */
+        /** The value which will replace the token in the filtering operation. */
         String value;
+        // CheckStyle:VisibilityModifier ON
 
         /**
-         * Constructor for the Filter object
+         * Constructor for the Filter object.
          *
-         * @param token  The token which will be replaced when filtering
-         * @param value  The value which will replace the token when filtering
+         * @param token  The token which will be replaced when filtering.
+         * @param value  The value which will replace the token when filtering.
          */
         public Filter(String token, String value) {
-           this.token = token;
-           this.value = value;
+           setToken(token);
+           setValue(value);
         }
 
         /**
-         * No argument conmstructor
+         * No-argument conmstructor.
          */
         public Filter() {
         }
 
         /**
-         * Sets the Token attribute of the Filter object
+         * Sets the Token attribute of the Filter object.
          *
-         * @param token  The new Token value
+         * @param token  The new Token value.
          */
         public void setToken(String token) {
            this.token = token;
         }
 
         /**
-         * Sets the Value attribute of the Filter object
+         * Sets the Value attribute of the Filter object.
          *
-         * @param value  The new Value value
+         * @param value  The new Value value.
          */
         public void setValue(String value) {
            this.value = value;
         }
 
         /**
-         * Gets the Token attribute of the Filter object
+         * Gets the Token attribute of the Filter object.
          *
-         * @return   The Token value
+         * @return   The Token value.
          */
         public String getToken() {
            return token;
         }
 
         /**
-         * Gets the Value attribute of the Filter object
+         * Gets the Value attribute of the Filter object.
          *
-         * @return   The Value value
+         * @return   The Value value.
          */
         public String getValue() {
            return value;
@@ -108,7 +110,7 @@ public class FilterSet extends DataType implements Cloneable {
     public class FiltersFile {
 
         /**
-         * Constructor for the Filter object
+         * Constructor for the FiltersFile object.
          */
         public FiltersFile() {
         }
@@ -119,7 +121,48 @@ public class FilterSet extends DataType implements Cloneable {
          * @param file the file from which filters will be read.
          */
         public void setFile(File file) {
-           readFiltersFromFile(file);
+           filtersFiles.add(file);
+        }
+    }
+
+    /**
+     * EnumeratedAttribute to set behavior WRT missing filtersfiles:
+     * "fail" (default), "warn", "ignore".
+     * @since Ant 1.7
+     */
+    public static class OnMissing extends EnumeratedAttribute {
+        private static final String[] VALUES
+            = new String[] {"fail", "warn", "ignore"};
+
+        /** Fail value */
+        public static final OnMissing FAIL = new OnMissing("fail");
+        /** Warn value */
+        public static final OnMissing WARN = new OnMissing("warn");
+        /** Ignore value */
+        public static final OnMissing IGNORE = new OnMissing("ignore");
+
+        private static final int FAIL_INDEX = 0;
+        private static final int WARN_INDEX = 1;
+        private static final int IGNORE_INDEX = 2;
+
+        /**
+         * Default constructor.
+         */
+        public OnMissing() {
+        }
+
+        /**
+         * Convenience constructor.
+         * @param value the value to set.
+         */
+        public OnMissing(String value) {
+            setValue(value);
+        }
+
+        //inherit doc
+        /** {@inheritDoc}. */
+        public String[] getValues() {
+            return VALUES;
         }
     }
 
@@ -132,19 +175,32 @@ public class FilterSet extends DataType implements Cloneable {
     private String startOfToken = DEFAULT_TOKEN_START;
     private String endOfToken = DEFAULT_TOKEN_END;
 
+    /** Contains a list of parsed tokens */
+    private Vector passedTokens;
+    /** if a duplicate token is found, this is set to true */
+    private boolean duplicateToken = false;
+
+    private boolean recurse = true;
+    private Hashtable filterHash = null;
+    private Vector filtersFiles = new Vector();
+    private OnMissing onMissingFiltersFile = OnMissing.FAIL;
+    private boolean readingFiles = false;
+
+    private int recurseDepth = 0;
+
     /**
      * List of ordered filters and filter files.
      */
     private Vector filters = new Vector();
 
     /**
-     * Default constructor
+     * Default constructor.
      */
     public FilterSet() {
     }
 
     /**
-     * Create a Filterset from another filterset
+     * Create a Filterset from another filterset.
      *
      * @param filterset the filterset upon which this filterset will be based.
      */
@@ -154,19 +210,29 @@ public class FilterSet extends DataType implements Cloneable {
     }
 
     /**
-     * Get the filters in the filter set
+     * Get the filters in the filter set.
      *
-     * @return a Vector of Filter instances
+     * @return a Vector of Filter instances.
      */
-    protected Vector getFilters() {
+    protected synchronized Vector getFilters() {
         if (isReference()) {
             return getRef().getFilters();
         }
+        dieOnCircularReference();
+        //silly hack to avoid stack overflow...
+        if (!readingFiles) {
+            readingFiles = true;
+            for (int i = 0, sz = filtersFiles.size(); i < sz; i++) {
+                readFiltersFromFile((File) filtersFiles.get(i));
+            }
+            filtersFiles.clear();
+            readingFiles = false;
+        }
         return filters;
     }
 
     /**
-     * Get the referred filter set
+     * Get the referenced filter set.
      *
      * @return the filterset from the reference.
      */
@@ -179,33 +245,39 @@ public class FilterSet extends DataType implements Cloneable {
      *
      * @return   The hash of the tokens and values for quick lookup.
      */
-    public Hashtable getFilterHash() {
-        int filterSize = getFilters().size();
-        Hashtable filterHash = new Hashtable(filterSize + 1);
-        for (Enumeration e = getFilters().elements(); e.hasMoreElements();) {
-           Filter filter = (Filter) e.nextElement();
-           filterHash.put(filter.getToken(), filter.getValue());
+    public synchronized Hashtable getFilterHash() {
+        if (isReference()) {
+            return getRef().getFilterHash();
+        }
+        dieOnCircularReference();
+        if (filterHash == null) {
+            filterHash = new Hashtable(getFilters().size());
+            for (Enumeration e = getFilters().elements(); e.hasMoreElements();) {
+               Filter filter = (Filter) e.nextElement();
+               filterHash.put(filter.getToken(), filter.getValue());
+            }
         }
         return filterHash;
     }
 
     /**
-     * set the file containing the filters for this filterset.
+     * Set the file containing the filters for this filterset.
      *
-     * @param filtersFile sets the filter fil to read filters for this filter set from.
-     * @exception BuildException if there is a problem reading the filters
+     * @param filtersFile sets the filter file from which to read filters
+     *        for this filter set.
+     * @throws BuildException if there is an error.
      */
     public void setFiltersfile(File filtersFile) throws BuildException {
         if (isReference()) {
             throw tooManyAttributes();
         }
-        readFiltersFromFile(filtersFile);
+        filtersFiles.add(filtersFile);
     }
 
     /**
-     * The string used to id the beginning of a token.
+     * Set the string used to id the beginning of a token.
      *
-     * @param startOfToken  The new Begintoken value
+     * @param startOfToken  The new Begintoken value.
      */
     public void setBeginToken(String startOfToken) {
         if (isReference()) {
@@ -218,9 +290,9 @@ public class FilterSet extends DataType implements Cloneable {
     }
 
     /**
-     * Get the begin token for this filterset
+     * Get the begin token for this filterset.
      *
-     * @return the filter set's begin token for filtering
+     * @return the filter set's begin token for filtering.
      */
     public String getBeginToken() {
         if (isReference()) {
@@ -229,11 +301,10 @@ public class FilterSet extends DataType implements Cloneable {
         return startOfToken;
     }
 
-
     /**
-     * The string used to id the end of a token.
+     * Set the string used to id the end of a token.
      *
-     * @param endOfToken  The new Endtoken value
+     * @param endOfToken  The new Endtoken value.
      */
     public void setEndToken(String endOfToken) {
         if (isReference()) {
@@ -246,9 +317,9 @@ public class FilterSet extends DataType implements Cloneable {
     }
 
     /**
-     * Get the end token for this filterset
+     * Get the end token for this filterset.
      *
-     * @return the filter set's end token for replacement delimiting
+     * @return the filter set's end token for replacement delimiting.
      */
     public String getEndToken() {
         if (isReference()) {
@@ -257,24 +328,36 @@ public class FilterSet extends DataType implements Cloneable {
         return endOfToken;
     }
 
+    /**
+     * Set whether recursive token expansion is enabled.
+     * @param recurse <code>boolean</code> whether to recurse.
+     */
+    public void setRecurse(boolean recurse) {
+        this.recurse = recurse;
+    }
+
+    /**
+     * Get whether recursive token expansion is enabled.
+     * @return <code>boolean</code> whether enabled.
+     */
+    public boolean isRecurse() {
+        return recurse;
+    }
 
     /**
      * Read the filters from the given file.
      *
-     * @param filtersFile         the file from which filters are read
-     * @exception BuildException  Throw a build exception when unable to read the
-     * file.
+     * @param filtersFile        the file from which filters are read.
+     * @exception BuildException when the file cannot be read.
      */
-    public void readFiltersFromFile(File filtersFile) throws BuildException {
+    public synchronized void readFiltersFromFile(File filtersFile) throws BuildException {
         if (isReference()) {
             throw tooManyAttributes();
         }
-
         if (!filtersFile.exists()) {
-            throw new BuildException("Could not read filters from file "
+           handleMissingFile("Could not read filters from file "
                                      + filtersFile + " as it doesn't exist.");
         }
-
         if (filtersFile.isFile()) {
            log("Reading filters from " + filtersFile, Project.MSG_VERBOSE);
            FileInputStream in = null;
@@ -284,28 +367,24 @@ public class FilterSet extends DataType implements Cloneable {
               props.load(in);
 
               Enumeration e = props.propertyNames();
-              Vector filters = getFilters();
+              Vector filts = getFilters();
               while (e.hasMoreElements()) {
                  String strPropName = (String) e.nextElement();
                  String strValue = props.getProperty(strPropName);
-                 filters.addElement(new Filter(strPropName, strValue));
+                 filts.addElement(new Filter(strPropName, strValue));
               }
            } catch (Exception ex) {
               throw new BuildException("Could not read filters from file: "
-                + filtersFile);
+                  + filtersFile, ex);
            } finally {
-              if (in != null) {
-                 try {
-                    in.close();
-                 } catch (IOException ioex) {
-                     // ignore
-                 }
-              }
+              FileUtils.close(in);
            }
         } else {
-           throw new BuildException("Must specify a file not a directory in "
-            + "the filtersfile attribute:" + filtersFile);
+           handleMissingFile(
+               "Must specify a file rather than a directory in "
+               + "the filtersfile attribute:" + filtersFile);
         }
+        filterHash = null;
     }
 
     /**
@@ -315,14 +394,111 @@ public class FilterSet extends DataType implements Cloneable {
      * This resets the passedTokens and calls iReplaceTokens to
      * do the actual replacements.
      *
-     * @param line  The line to process the tokens in.
-     * @return      The string with the tokens replaced.
+     * @param line  The line in which to process embedded tokens.
+     * @return      The input string after token replacement.
      */
-    public String replaceTokens(String line) {
-        passedTokens = null; // reset for new line
+    public synchronized String replaceTokens(String line) {
         return iReplaceTokens(line);
     }
 
+    /**
+     * Add a new filter.
+     *
+     * @param filter the filter to be added.
+     */
+    public synchronized void addFilter(Filter filter) {
+        if (isReference()) {
+            throw noChildrenAllowed();
+        }
+        filters.addElement(filter);
+        filterHash = null;
+    }
+
+    /**
+     * Create a new FiltersFile.
+     *
+     * @return The filtersfile that was created.
+     */
+    public FiltersFile createFiltersfile() {
+        if (isReference()) {
+            throw noChildrenAllowed();
+        }
+        return new FiltersFile();
+    }
+
+    /**
+     * Add a new filter made from the given token and value.
+     *
+     * @param token The token for the new filter.
+     * @param value The value for the new filter.
+     */
+    public synchronized void addFilter(String token, String value) {
+        if (isReference()) {
+            throw noChildrenAllowed();
+        }
+        addFilter(new Filter(token, value));
+    }
+
+    /**
+     * Add a Filterset to this filter set.
+     *
+     * @param filterSet the filterset to be added to this filterset
+     */
+    public synchronized void addConfiguredFilterSet(FilterSet filterSet) {
+        if (isReference()) {
+            throw noChildrenAllowed();
+        }
+        for (Enumeration e = filterSet.getFilters().elements(); e.hasMoreElements();) {
+            addFilter((Filter) e.nextElement());
+        }
+    }
+
+    /**
+    * Test to see if this filter set has filters.
+    *
+    * @return Return true if there are filters in this set.
+    */
+    public synchronized boolean hasFilters() {
+        return getFilters().size() > 0;
+    }
+
+    /**
+     * Clone the filterset.
+     *
+     * @return a deep clone of this filterset.
+     *
+     * @throws BuildException if the clone cannot be performed.
+     */
+    public synchronized Object clone() throws BuildException {
+        if (isReference()) {
+            return ((FilterSet) getRef()).clone();
+        }
+        try {
+            FilterSet fs = (FilterSet) super.clone();
+            fs.filters = (Vector) getFilters().clone();
+            fs.setProject(getProject());
+            return fs;
+        } catch (CloneNotSupportedException e) {
+            throw new BuildException(e);
+        }
+    }
+
+    /**
+     * Set the behavior WRT missing filtersfiles.
+     * @param onMissingFiltersFile the OnMissing describing the behavior.
+     */
+    public void setOnMissingFiltersFile(OnMissing onMissingFiltersFile) {
+        this.onMissingFiltersFile = onMissingFiltersFile;
+    }
+
+    /**
+     * Get the onMissingFiltersFile setting.
+     * @return the OnMissing instance.
+     */
+    public OnMissing getOnMissingFiltersFile() {
+        return onMissingFiltersFile;
+    }
+
     /**
      * Does replacement on the given string with token matching.
      * This uses the defined begintoken and endtoken values which default
@@ -331,7 +507,7 @@ public class FilterSet extends DataType implements Cloneable {
      * @param line  The line to process the tokens in.
      * @return      The string with the tokens replaced.
      */
-    private String iReplaceTokens(String line) {
+    private synchronized String iReplaceTokens(String line) {
         String beginToken = getBeginToken();
         String endToken = getEndToken();
         int index = line.indexOf(beginToken);
@@ -344,7 +520,8 @@ public class FilterSet extends DataType implements Cloneable {
                 String token = null;
                 String value = null;
 
-                do {
+                while (index > -1) {
+                    //can't have zero-length token
                     int endIndex = line.indexOf(endToken,
                         index + beginToken.length() + 1);
                     if (endIndex == -1) {
@@ -355,7 +532,7 @@ public class FilterSet extends DataType implements Cloneable {
                     b.append(line.substring(i, index));
                     if (tokens.containsKey(token)) {
                         value = (String) tokens.get(token);
-                        if (!value.equals(token)) {
+                        if (recurse && !value.equals(token)) {
                             // we have another token, let's parse it.
                             value = replaceTokens(value, token);
                         }
@@ -365,11 +542,16 @@ public class FilterSet extends DataType implements Cloneable {
                         i = index + beginToken.length() + token.length()
                             + endToken.length();
                     } else {
-                        // just append beginToken and search further
-                        b.append(beginToken);
-                        i = index + beginToken.length();
+                        // just append first character of beginToken
+                        // and search further
+                        // we can't skip the complete beginToken since
+                        // it may contain the start of another
+                        // candidate begin token (Bugzilla 45094)
+                        b.append(beginToken.charAt(0));
+                        i = index + 1;
                     }
-                } while ((index = line.indexOf(beginToken, i)) > -1);
+                    index = line.indexOf(beginToken, i);
+                }
 
                 b.append(line.substring(i));
                 return b.toString();
@@ -381,138 +563,65 @@ public class FilterSet extends DataType implements Cloneable {
         }
     }
 
-    /** Contains a list of parsed tokens */
-    private Vector passedTokens;
-    /** if a ducplicate token is found, this is set to true */
-    private boolean duplicateToken = false;
-
     /**
      * This parses tokens which point to tokens.
      * It also maintains a list of currently used tokens, so we cannot
-     * get into an infinite loop
-     * @param line the value / token to parse
-     * @param parent the parant token (= the token it was parsed from)
+     * get into an infinite loop.
+     * @param line the value / token to parse.
+     * @param parent the parent token (= the token it was parsed from).
      */
-    private String replaceTokens(String line, String parent)
+    private synchronized String replaceTokens(String line, String parent)
         throws BuildException {
-        if (passedTokens == null) {
+        String beginToken = getBeginToken();
+        String endToken = getEndToken();
+        if (recurseDepth == 0) {
             passedTokens = new Vector();
         }
+        recurseDepth++;
         if (passedTokens.contains(parent) && !duplicateToken) {
             duplicateToken = true;
-            StringBuffer sb = new StringBuffer();
-            sb.append("Infinite loop in tokens. Currently known tokens : ");
-            sb.append(passedTokens);
-            sb.append("\nProblem token : " + getBeginToken() + parent
-                + getEndToken());
-            sb.append(" called from " + getBeginToken()
-                + passedTokens.lastElement());
-            sb.append(getEndToken());
-            System.out.println(sb.toString());
+            System.out.println(
+                "Infinite loop in tokens. Currently known tokens : "
+                + passedTokens.toString() + "\nProblem token : " + beginToken
+                + parent + endToken + " called from " + beginToken
+                + passedTokens.lastElement().toString() + endToken);
+            recurseDepth--;
             return parent;
         }
         passedTokens.addElement(parent);
         String value = iReplaceTokens(line);
-        if (value.indexOf(getBeginToken()) == -1 && !duplicateToken) {
-            duplicateToken = false;
+        if (value.indexOf(beginToken) == -1 && !duplicateToken
+                && recurseDepth == 1) {
             passedTokens = null;
         } else if (duplicateToken) {
             // should always be the case...
             if (passedTokens.size() > 0) {
-                value = (String) passedTokens.lastElement();
-                passedTokens.removeElementAt(passedTokens.size() - 1);
+                value = (String) passedTokens.remove(passedTokens.size() - 1);
                 if (passedTokens.size() == 0) {
-                    value = getBeginToken() + value + getEndToken();
+                    value = beginToken + value + endToken;
                     duplicateToken = false;
                 }
             }
+        } else if (passedTokens.size() > 0) {
+            // remove last seen token when crawling out of recursion 
+            passedTokens.remove(passedTokens.size() - 1);
         }
+        recurseDepth--;
         return value;
     }
 
-    /**
-     * Create a new filter
-     *
-     * @param filter the filter to be added
-     */
-    public void addFilter(Filter filter) {
-        if (isReference()) {
-            throw noChildrenAllowed();
-        }
-        filters.addElement(filter);
-    }
-
-    /**
-     * Create a new FiltersFile
-     *
-     * @return   The filter that was created.
-     */
-    public FiltersFile createFiltersfile() {
-        if (isReference()) {
-            throw noChildrenAllowed();
-        }
-        return new FiltersFile();
-    }
-
-    /**
-    * Add a new filter made from the given token and value.
-    *
-    * @param token  The token for the new filter.
-    * @param value  The value for the new filter.
-    */
-    public void addFilter(String token, String value) {
-        if (isReference()) {
-            throw noChildrenAllowed();
-        }
-        filters.addElement(new Filter(token, value));
-    }
-
-    /**
-    * Add a Filterset to this filter set
-    *
-    * @param filterSet the filterset to be added to this filterset
-    */
-    public void addConfiguredFilterSet(FilterSet filterSet) {
-        if (isReference()) {
-            throw noChildrenAllowed();
-        }
-        for (Enumeration e = filterSet.getFilters().elements(); e.hasMoreElements();) {
-            filters.addElement(e.nextElement());
-        }
-    }
-
-    /**
-    * Test to see if this filter set it empty.
-    *
-    * @return   Return true if there are filter in this set otherwise false.
-    */
-    public boolean hasFilters() {
-        return getFilters().size() > 0;
-    }
-
-    /**
-     * clone the filterset
-     *
-     * @return a deep clone of this filterset
-     *
-     * @throws BuildException if the clone cannot be performed.
-     */
-    public Object clone() throws BuildException {
-        if (isReference()) {
-            return ((FilterSet) getRef()).clone();
-        } else {
-            try {
-                FilterSet fs = (FilterSet) super.clone();
-                fs.filters = (Vector) getFilters().clone();
-                fs.setProject(getProject());
-                return fs;
-            } catch (CloneNotSupportedException e) {
-                throw new BuildException(e);
-            }
+    private void handleMissingFile(String message) {
+        switch (onMissingFiltersFile.getIndex()) {
+        case OnMissing.IGNORE_INDEX:
+            return;
+        case OnMissing.FAIL_INDEX:
+            throw new BuildException(message);
+        case OnMissing.WARN_INDEX:
+            log(message, Project.MSG_WARN);
+            return;
+        default:
+            throw new BuildException("Invalid value for onMissingFiltersFile");
         }
     }
 
 }
-
-
-
